/* Copyright (C) 2002 Jacob Kroon, see COPYING for details */

#include <stdlib.h>
#include <unistd.h>
#include <vorbis/vorbisfile.h>
#include "nogger.h"

/* defines */
#define BUFFER_SIZE		4096

/* static variables */
static int _play = 0;
static int _jump = 0;
static int _pause = 0;
static int _random = 0;
static double _jump_time = 0;
static double _elapsed_time = 0;
static double _total_time = 0;
static GCond *_jump_cond = NULL;
static GCond *_stop_cond = NULL;
static GStaticMutex _mutex = G_STATIC_MUTEX_INIT;
static vorbis_info *_info = NULL;

/* static functions */
static gpointer _thread_func(gpointer filename)
{
  int loop = 1, ret_value;
  long bytes_read = -1;
  unsigned char *buffer = NULL;
  FILE *file = NULL;
  OggVorbis_File vf;

  g_static_mutex_lock(&_mutex);
  _jump_cond = g_cond_new();
  _stop_cond = g_cond_new();
  buffer = malloc(BUFFER_SIZE);
  if ((file = fopen(filename, "r")) == NULL)
    goto done;
  g_free(filename);
  if (ov_open(file, &vf, NULL, 0) < 0)
    goto done;
  _info = ov_info(&vf, -1);
  _elapsed_time = 0;
  _total_time = ov_time_total(&vf, -1);
  _play = 1;
  audio_open(_info->rate, _info->channels);
  g_static_mutex_unlock(&_mutex);
  gdk_threads_enter();
  gui_update(NULL);
  gdk_threads_leave();
  do {
    if (!_play)
      loop = 0;
    if (_jump) {
      ov_time_seek(&vf, _jump_time);
      _jump = 0;
      _elapsed_time = _jump_time;
      g_cond_signal(_jump_cond);
    } else
      _elapsed_time = ov_time_tell(&vf);
    if (_pause) {
      usleep(100000);
      continue;
    }
    bytes_read = ov_read(&vf, buffer, BUFFER_SIZE, 0, 2, 1, NULL);
    g_static_mutex_lock(&_mutex);
    audio_write(buffer, bytes_read);
    g_static_mutex_unlock(&_mutex);
  } while (loop && bytes_read != 0);
  g_static_mutex_lock(&_mutex);
  audio_close();
  g_static_mutex_unlock(&_mutex);
  ov_clear(&vf);
  file = NULL;
  _play = 0;
done:
  if (file != NULL)
    fclose(file);
  if (buffer != NULL)
    free(buffer);
  g_cond_free(_jump_cond);
  _jump_cond = NULL;
  if (loop && bytes_read == 0) {
    if (_random)
      ret_value = -1;
    else
      ret_value = 1;
  } else
    ret_value = 0;
  g_cond_signal(_stop_cond);
  return (gpointer) ret_value;
}

/* functions */
GThread *ogg_play(char *filename)
{
  GMutex *pmutex;

  if (_play) {
    g_static_mutex_lock(&_mutex);
    _play = 0;
    pmutex = g_static_mutex_get_mutex(&_mutex);
    if (_stop_cond != NULL) {
      g_cond_wait(_stop_cond, pmutex);
      g_cond_free(_stop_cond);
      _stop_cond = NULL;
    }
    g_static_mutex_unlock(&_mutex);
  }
  return g_thread_create(_thread_func, filename, TRUE, NULL);
}

void ogg_toggle_pause(void)
{
  g_static_mutex_lock(&_mutex);
  if (_pause) {
    _pause = 0;
    audio_open(_info->rate, _info->channels);
  } else {
    _pause = 1;
    audio_close();
  }
  g_static_mutex_unlock(&_mutex);
}

void ogg_toggle_random(void)
{
  if (_random)
    _random = 0;
  else
    _random = 1;
}

void ogg_get_times(int *elapsed_time, int *total_time)
{
  *elapsed_time = (int) _elapsed_time;
  *total_time = (int) _total_time;
}

void ogg_jump_to(double jump_to)
{
  GMutex *pmutex;

  g_static_mutex_lock(&_mutex);
  if (_jump_cond == NULL) {
    g_static_mutex_unlock(&_mutex);
    return;
  }
  _jump = 1;
  _jump_time = jump_to * _total_time;
  pmutex = g_static_mutex_get_mutex(&_mutex);
  g_cond_wait(_jump_cond, pmutex);
  g_static_mutex_unlock(&_mutex);
}
