/* Copyright (C) 2002 Jacob Kroon, see COPYING for details */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include "nogger.h"

/* enums */
enum {
  SONG_COLUMN,
  FILENAME_COLUMN,
  NBR_OF_COLUMNS
};

/* static variables */
static int _nbr_of_files = 0;
static GtkTreeStore *_store = NULL;
static GtkTreeSelection *_selection = NULL;
static GtkTreeIter _iter;
static GThread *_ogg_thread = NULL;

/* static functions */
static void _trim_line(char *line)
{
  int i;

  for (i = 0; line[i] != 0; i++)
    if (line[i] == '_')
      line[i] = ' ';
}

static int _fill(char *dir_name, GtkTreeIter * parent_iter)
{
  int nbr_of_files_found = 0;
  char buffer[256], *file_name;
  DIR *dir;
  struct stat file_stat;
  struct dirent *dir_entry;
  GtkTreeIter *iter;

  if (lstat(dir_name, &file_stat) == -1 || !S_ISDIR(file_stat.st_mode)
      || (dir = opendir(dir_name)) == NULL) {
    if (parent_iter != NULL)
      gtk_tree_store_remove(_store, parent_iter);
    return 0;
  }
  iter = malloc(sizeof(*iter));
  while ((dir_entry = readdir(dir)) != NULL) {
    file_name = dir_entry->d_name;
    if (strcmp(file_name, ".") == 0 || strcmp(file_name, "..") == 0)
      continue;
    sprintf(buffer, "%s/%s", dir_name, file_name);
    _trim_line(file_name);
    lstat(buffer, &file_stat);
    if (S_ISDIR(file_stat.st_mode)) {
      gtk_tree_store_append(_store, iter, parent_iter);
      gtk_tree_store_set(_store, iter, SONG_COLUMN, file_name, -1);
      nbr_of_files_found += _fill(buffer, iter);
    } else if (strcmp(file_name + strlen(file_name) - 4, ".ogg") == 0) {
      file_name[strlen(file_name) - 4] = 0;
      gtk_tree_store_append(_store, iter, parent_iter);
      gtk_tree_store_set(_store, iter, SONG_COLUMN, file_name, -1);
      gtk_tree_store_set(_store, iter, FILENAME_COLUMN, buffer, -1);
      nbr_of_files_found++;
    }
  }
  if (nbr_of_files_found == 0)
    if (parent_iter != NULL)
      gtk_tree_store_remove(_store, parent_iter);
  closedir(dir);
  free(iter);
  return nbr_of_files_found;
}

static char *_get_filename(GtkTreeModel * model)
{
  char *filename;

  gtk_tree_model_get(model, &_iter, FILENAME_COLUMN, &filename, -1);
  return filename;
}

static void _select(int offset)
{
  int i;
  GtkTreeIter parent, temp_iter;
  GtkTreeModel *model;
  gchar *filename;

  if (offset < 0) {
    _select(1 + (int) ((float) random() / RAND_MAX * (_nbr_of_files - 1)));
    return;
  }
  gdk_threads_enter();
  gtk_tree_selection_get_selected(_selection, &model, NULL);
  gtk_tree_selection_unselect_all(_selection);
  gtk_tree_model_iter_parent(model, &parent, &_iter);
  for (i = 0; i < offset; i++) {
    temp_iter = _iter;
    if (gtk_tree_model_iter_has_child(model, &_iter)) {
      parent = _iter;
      gtk_tree_model_iter_nth_child(model, &_iter, &parent, 0);
    } else
      while (!gtk_tree_model_iter_next(model, &_iter)) {
	_iter = temp_iter;
	if (!gtk_tree_model_iter_parent(model, &parent, &_iter)) {
	  gtk_tree_model_get_iter_first(model, &_iter);
	  break;
	}
	temp_iter = _iter = parent;
      }
    if ((filename = _get_filename(model)) == NULL)
      i--;
    else
      g_free(filename);
  }
  gtk_tree_selection_select_iter(_selection, &_iter);
  if (!gtk_tree_selection_iter_is_selected(_selection, &_iter)) {
    filename = _get_filename(model);
    gdk_threads_leave();
    _ogg_thread = ogg_play(filename);
  } else
    gdk_threads_leave();
}

static void _changed_cb(GtkTreeSelection * selection, gpointer data)
{
  GtkTreeModel *model;
  gchar *filename;

  if (gtk_tree_selection_get_selected(selection, &model, &_iter)) {
    filename = _get_filename(model);
    if (filename != NULL)
      _ogg_thread = ogg_play(filename);
  }
}

static gpointer _thread_func(gpointer data)
{
  int next_song = 0;

  while (1) {
    usleep(10000);
    if (_ogg_thread != NULL) {
      next_song = (int) g_thread_join(_ogg_thread);
      if (next_song != 0)
	_select(next_song);
    }
  }
}

/* functions */
GtkWidget *tree_create(char *root_dir)
{
  GtkWidget *tree_window, *tree_view;
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;

  _store = gtk_tree_store_new(NBR_OF_COLUMNS, G_TYPE_STRING, G_TYPE_STRING);
  renderer = gtk_cell_renderer_text_new();
  column =
    gtk_tree_view_column_new_with_attributes(NULL, renderer, "text",
					     SONG_COLUMN, NULL);
  tree_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(_store));
  g_object_set(G_OBJECT(tree_view), "rules-hint", TRUE, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view), column);
  gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree_view), FALSE);
  gtk_widget_show(tree_view);
  _selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
  gtk_tree_selection_set_mode(_selection, GTK_SELECTION_SINGLE);
  g_signal_connect(G_OBJECT(_selection), "changed", G_CALLBACK(_changed_cb),
		   NULL);
  tree_window = gtk_scrolled_window_new(NULL, NULL);
  g_object_set(G_OBJECT(tree_window), "hscrollbar-policy", GTK_POLICY_NEVER,
	       NULL);
  g_object_set(G_OBJECT(tree_window), "vscrollbar-policy",
	       GTK_POLICY_AUTOMATIC, NULL);
  gtk_container_add(GTK_CONTAINER(tree_window), tree_view);
  g_thread_create(_thread_func, NULL, TRUE, NULL);
  if ((_nbr_of_files = _fill(root_dir, NULL)) == 0)
    return NULL;
  return tree_window;
}
